#include <stdio.h>
#include <stdlib.h>
#include <strings.h>

#include "election_def.h"

/*******************************************************************************
*
*   PROTOTYPES
*
********************************************************************************/

int isSomeoneJustExcluded ( electionDef *myE);
int isSomeoneJustElected ( electionDef *myE);
void distribute_elect ( electionDef *myE, int thisWinner, int surplus);
int distribute_exclude ( electionDef *myE, int thisWinner);
int num_remaining_seats ( electionDef *myE);
int num_remaining_cand ( electionDef *myE);
int voteCount (candidateDef *cand);
int voteCountUpTo (candidateDef *cand, int upTo);
int electCandidate( electionDef *myE, int thisWinner);
int candParty2Num( electionDef *myE, voteDef name);
int sortPrefs (const void *in1, const void *in2) ;
void sortByPrefs(electionDef *myE, int prefNum);
void add_vote(electionDef *myE, int cand, int numVotes, tValDef tVal, ballotList theVotes, int countNum);
int ceaseScrutiny(electionDef *myE);
int pickExcludedCandToDist(electionDef *myE);
int pickElectedCandToDist(electionDef *myE);
int pickCandToExclude(electionDef *myE);
void freeElection(electionDef *myE);

/****************************************************************/
/*   MAIN */
/****************************************************************/
int main (int argc, const char * argv[]) {
    int i;
    electionDef myElection;
    int electionDone, someoneElected, doExclusion;
    int distCount[MAXCAND+1];
    tValDef tmpTVal;
#ifdef DEBUG_PRINT
    int j;
#endif
          
    printf("c_hareclarke_counter - compiled %s.\n<br>\n<br>", __DATE__);

    /* STEP 0a:  Initialise the election structure */
    myElection.nextElectedCandidate = 0;
    myElection.vacancyCount = 0;
    myElection.totalBallotCount = 0;
    myElection.totalCandidateCount = 0;
    myElection.totalPartyCount = 0;
    myElection.totalElectorateCount = 0;
    myElection.candfile = myElection.partyfile = myElection.ballotfile = NULL;
    myElection.ecode = 0;
    myElection.countNum = 0;
    myElection.exhaustedBallots = NULL;
    
    /* STEP 0b.  Read in our votes, and the candidates.*/
    /* If there is an error reading input, print help and exit */
    if (read_cmdline(argc, argv, &myElection) != 0) {
        print_help();
        return -1;
    }

    /* Ask which electorate we want to do the election for... */
    if (myElection.totalElectorateCount == 1) {
        myElection.ecode = 0;
    }
    else if (myElection.totalElectorateCount < 2) {
        error_exit("Error: bad electorate count (%d)\n", myElection.totalElectorateCount);
        
    }
    else {
        fprintf(stderr, "Chose an electorate:\n");

        for (i = 0 ; i < myElection.totalElectorateCount ; i++) {
            fprintf(stderr, "%d : %s\n", i, myElection.electorates[i].name);
        }
        myElection.ecode = myElection.electorates[getIntegerFromUser(0, myElection.totalElectorateCount-1)].ecode;
    }
        
    fprintf(stderr, "Running an election for %s\n",
           myElection.electorates[myElection.ecode].name);
    
    printf("Running an election for %s\n<br>",
           myElection.electorates[myElection.ecode].name);
    
    /* Set the vacancy count now that we know the electorate number */
    myElection.vacancyCount = myElection.electorates[myElection.ecode].seats;
    
    /* Now read in the votes, candidates, parties for our chosen election */    
    read_parties(&myElection, myElection.partyfile);
    read_candidates(&myElection, myElection.candfile);
    
    /* Sort the candidates alphabetically within parties - this is to make the output similar to evacs */
    /* Store the sorted candidate in the alphaCandidates pointer in the electionDef structure */
    sort_candidates_alpha(&myElection);
        
#ifdef DEBUG_PRINT
    for (i = 0 ; i < myElection.totalPartyCount ; i++) {

        printf("Party %d:  %s : %d candidates\n<br>", myElection.parties[i].pcode,  myElection.parties[i].name,  myElection.parties[i].numcand);

        for (j = 0 ; j <  myElection.parties[i].numcand ; j++) {
            printf("\t%d %s\n<br>", myElection.candidates[myElection.parties[i].candidates[j]].ccode, myElection.candidates[myElection.parties[i].candidates[j]].name);
        }        
    }
#endif

    /* If we have enough seats for all potetial candidates, they all win */
    /* see section 111 of legislation.  We don't even need to do a count of votes.  (Actually, there shouldn't even
        be any votes, as there doesn't even need to be a poll done).  */
    if(myElection.vacancyCount >= myElection.totalCandidateCount) {        
        for(i = 0; i < myElection.totalCandidateCount ; i++) {
            electCandidate(&myElection, i);
        }
        
        printf("Enough candidates for the vacancies - everyone is elected.\n");
        
        printStart(&myElection);
        printEnd(&myElection, 1);

        return 0;
    }
    
    /* Need to do a poll under s111, so read in votes */
    read_ballots(&myElection, myElection.ballotfile);
    
    /* We should now have the election structure filled out with number of votes,
      candidates, and arrays to the ballots and candidates.  We should also
        know the vacancy count.   Informal votes have been removed, as have dead candidates.  :-) */

#ifdef DEBUG_PRINT
    printf("ballots %d, cands %d, vac %d\n<br>", myElection.totalBallotCount,
           myElection.totalCandidateCount,
           myElection.vacancyCount);
#endif
    
    /* STEP 0c:  Some sanity checking. */
    
    /* If no candidates, or no ballots, or no vacancies, exit */
    if (myElection.totalBallotCount < 1 ||
        myElection.totalCandidateCount < 1 ||
        myElection.vacancyCount < 1 ||
        myElection.totalPartyCount < 1) {

        printf("\nNot enough input!  Cowardly quiting.\n");
        print_help();
        return 0;        
    }

    printf("Sanity checks OK\n<br>");
    fflush(stdout);
    
    /* STEP 1:  count first preference votes. */
    myElection.countNum = 1;
    
    for (i = 0 ; i < TABLE_SIZE ; i++) {
        int candNum;
        ballotList testBallot;
        ballotList tmpBallot;
                
        testBallot = myElection.ballots[i];
        
        while (testBallot != EMPTY_LIST) {

            candNum = candParty2Num(&myElection,
                                testBallot->orderedVotes[1]);

            tmpBallot = testBallot;
            testBallot = testBallot->next;
            tmpBallot->next = NULL;
            
           // printf("add a vote to %d\n", candNum);
            
            tmpTVal.numerator = 1;
            tmpTVal.denominator = 1;
            
            add_vote(&myElection, candNum, 1, tmpTVal, tmpBallot, myElection.countNum);        
        }        
    }

    printf("Votes added: %d formal votes\n<br>", myElection.totalBallotCount);
    fflush(stdout);
    
#ifdef DEBUG_PRINT
    for (i = 0 ; i < myElection.totalCandidateCount ; i++) {

        printf("Candidate %d: %s : %d 1st pref votes\n<br>", i,
               myElection.candidates[i].name,
               voteCount(&(myElection.candidates[i])));
    }
    fflush(stdout);
#endif    
    
    /* STEP 2:  calculate the quota for the election. */

    myElection.quota = myElection.totalBallotCount /
        (myElection.vacancyCount + 1) + 1;

    printf("Quota:  (%d / (%d + 1)) + 1 = %d\n<br>\n<br>", myElection.totalBallotCount, myElection.vacancyCount, 
           myElection.quota);

    electionDone = 0;

    printStart(&myElection);
    
    for (i = myElection.totalCandidateCount ; i < MAXCAND+1 ; i++) {
        distCount[i] = 0;
    }
    for (i = 0 ; i < myElection.totalCandidateCount; i++) {
        distCount[i] = voteCount(&(myElection.candidates[i]));
    }
    
    tmpTVal.numerator = 1;
    tmpTVal.denominator = 1;
    
    printRound(&myElection, myElection.countNum, distCount, 0, tmpTVal);

    /*******************************************************************/
    /* We now loop through steps 3, 4 and 5 until the election is done. */
    /*******************************************************************/
    while (! electionDone) {

        someoneElected = 0;
        doExclusion = 1;

        myElection.countNum ++;
        
#ifdef DEBUG_PRINT
        for (i = 0; i < myElection.totalCandidateCount ; i++) {

            printf("Candidate %d: %s : %d votes\n",
                   myElection.orderedCandidates[i].index,
                   myElection.orderedCandidates[i].name,
                   voteCount(&(myElection.orderedCandidates[i])));

        }
#endif        

        /* STEP 3:  Elect first person over quota */

        for (i = 0; i < myElection.totalCandidateCount ; i++) {

            if (voteCount(&(myElection.candidates[i])) >= myElection.quota &&
                myElection.candidates[i].status == limbo) {

               // printf("Will elect someone over quota...\n");
                electCandidate(&myElection, i);
                
            }
        }

        if (ceaseScrutiny(&myElection)) {
            electionDone = 1;
        }
        
        /* STEP 4: Distribute if someone elected, and noone is being excluded */
        
        if (isSomeoneJustElected(&myElection) &&
            isSomeoneJustExcluded(&myElection) == 0 && ! electionDone) {
            int candToDist;
            
            //   printf("Someone is justelected\n");
            candToDist = pickElectedCandToDist(&myElection);
            // printf("Cand To DIST %d\n", candToDist);
            
            if (candToDist > -1) {
                if (myElection.candidates[candToDist].surplus > 0) {
                    myElection.candidates[candToDist].status = elected;
                    //      printf("Call distribute_elect %d\n", myElection.orderedCandidates[i].index);
                    
                    distribute_elect(&myElection, candToDist, myElection.candidates[candToDist].surplus);
                    doExclusion = 0;
                }
                else {
                    myElection.candidates[candToDist].status = elected;
                }
            }
        }

        if (ceaseScrutiny(&myElection)) {
            electionDone = 1;
        }
        
        /* STEP 5:  Exclude someone if no-one over quota, and noone already being excluded */

        if (doExclusion &&
                isSomeoneJustExcluded(&myElection) == 0 && ! electionDone) {
            int candToDist;
            
            //            printf("DO exclusion\n");
            
            if (num_remaining_cand(&myElection) >
                num_remaining_seats(&myElection)) {

                candToDist = pickCandToExclude(&myElection); 
               
                if (candToDist > -1) {
                    /*   printf("Exluding %d (%s)\n",
                    myElection.orderedCandidates[i].index,
                    myElection.orderedCandidates[i].name);  */

                    myElection.candidates[candToDist].status = justexcluded;
                    myElection.candidates[candToDist].statusObtainedAtCount = myElection.countNum;
                }
            }
        }
        
        if (ceaseScrutiny(&myElection)) {
            electionDone = 1;
        }
        

        /* Process next step of current exclusion */
        if (doExclusion && isSomeoneJustExcluded(&myElection) &&  ! electionDone) {
            int doneDistrib = 0;
            int candToDist;
            
            //  printf("Someone is justexcluded\n");

            candToDist = pickExcludedCandToDist(&myElection);            

            if (candToDist > -1) {
            
                doneDistrib = distribute_exclude(&myElection, candToDist);

                if (doneDistrib == 0) {
                    myElection.candidates[candToDist].status = excluded;
                }
            }
        }

        if (ceaseScrutiny(&myElection)) {
            electionDone = 1;
        }
                
       // printf("End of loop. Loop back\n");
      //  fflush(stdout);        
    }  // end of while(electionNotDone) loop

    printEnd(&myElection,  myElection.countNum + 1);

    printf("<P>Elected Candidates are:  <br>\n");
    
    for (i = 0 ; i < myElection.nextElectedCandidate ; i++) {
     
        printf("%d: %s (%s)\n<br>", i+1, myElection.candidates[myElection.electedCandidates[i]].name,
               myElection.parties[myElection.candidates[myElection.electedCandidates[i]].pcode].abbrev);
        
    }
    
    freeElection(&myElection);
    
    return 0;
}

/******************************************************************************/
int voteCount (candidateDef *cand) {
    int total = 0;
    weightedVoteDef *a = cand->voteCount;

    while (a != NULL) {
        total += (float)a->numVotes * ((float)a->tVal.numerator / (float)a->tVal.denominator);
        a = a->next;
    }
    
    return total;
}

int voteCountUpTo (candidateDef *cand, int upTo) {
    int total = 0;
    weightedVoteDef *a = cand->voteCount;
    
    while (a != NULL) {
        
        if (a->countNum <= upTo) {
            total += (float)a->numVotes * ((float)a->tVal.numerator / (float)a->tVal.denominator);
        }
        a = a->next;
    }
    
    return total;
}

/******************************************************************************/
void add_vote(electionDef *myE, int cand, int numVotes, tValDef tVal, ballotList theVotes, int countNum) {
    

    weightedVoteDef *a = myE->candidates[cand].voteCount;
    weightedVoteDef *prev = NULL;
    weightedVoteDef *tmp;
    ballotList tmplist;
    
  //  printf("Add to %d\n", countNum);
    
    while (a != NULL && tVal_gt(a->tVal,tVal)) {
        a = a->next;
        prev = a;
    }

    while (a != NULL && a->countNum < countNum) {
        a = a->next;
        prev = a;
    }
        
    if (a == NULL && prev == NULL) {
        /* New start to list needed */
    
        tmp = malloc (sizeof(weightedVoteDef));
        tmp->numVotes = numVotes;
        tmp->tVal.numerator = tVal.numerator;
        tmp->tVal.denominator = tVal.denominator;
        tmp->next = myE->candidates[cand].voteCount;
        tmp->votelist = theVotes;
        tmp->countNum = countNum;
        
        myE->candidates[cand].voteCount = tmp;

    }
    else if (a == NULL) {
        /* New end of list needed */
        
        tmp = malloc (sizeof(weightedVoteDef));
        tmp->numVotes = numVotes;
        tmp->tVal.numerator = tVal.numerator;
        tmp->tVal.denominator = tVal.denominator;
        tmp->next = prev->next;
        tmp->votelist = theVotes;
        tmp->countNum = countNum;
           
        prev->next = tmp;
                
    }
    else if (tVal_e(a->tVal,tVal) && a->countNum == countNum) {
        /* Need to update this 'group' of ballots */

        a->numVotes += numVotes;
        tmplist = theVotes;
        
        while (tmplist->next != NULL) {
            tmplist = tmplist->next;
        }

        tmplist->next = a->votelist;
        a->votelist = theVotes;
        
    }
    else {
        /* a new group of ballots needed */

        tmp = malloc (sizeof(weightedVoteDef));
        tmp->numVotes = numVotes;
        tmp->tVal.numerator = tVal.numerator;
        tmp->tVal.denominator = tVal.denominator;
        tmp->next = prev->next;
        tmp->votelist = theVotes;
        tmp->countNum = countNum;
        
        prev->next = tmp;        

    }
}


/******************************************************************************/
/*  See schedule 4, Part 4.2, Clause 8, (1,2) */
int pickCandToExclude(electionDef *myE) {
    int i;        
    int ret = -1;
    int potentials[MAXCAND];
    int numPotentials = 1;
    int upToCount;
    
    for (i = 0; i < myE->totalCandidateCount ; i++) {
        if (myE->candidates[i].status == limbo) {
            if (ret == -1) {
                ret = i;                
            } else if (voteCount(&(myE->candidates[i])) < voteCount(&(myE->candidates[ret]))) {
                /* If least votes  s (8)(1)*/                
                ret = i;
            } else if (voteCount(&(myE->candidates[i])) == voteCount(&(myE->candidates[ret]))) {
                /*Same count number to obtain quota: (8)(2)*/
                                    
                printf("%d has hit same votes\n", i);
                /* Do countback here  (8)(2)(a) */
                
                upToCount = myE->countNum;
                while (voteCountUpTo(&myE->candidates[i], upToCount) == 
                       voteCountUpTo(&myE->candidates[ret], upToCount) && upToCount >= 0) {
                    
                    upToCount --;
                }
                
                if (voteCountUpTo(&myE->candidates[i], upToCount) == 
                    voteCountUpTo(&myE->candidates[ret], upToCount)) {
                    /* No difference in votes when counting back, so add this new voter as a potential
                    for the commissioner to choose by lot (8)(2)(b) */
                    
                    potentials[numPotentials] = i;
                    numPotentials++;
                } else if (voteCountUpTo(&myE->candidates[i], upToCount) < 
                           voteCountUpTo(&myE->candidates[ret], upToCount)) {
                    /* this candidate had less at the first unequal count - so they are now the one to be excluded */
                    ret = i;
                }
                
                
            }
        }
    }
    
    if (numPotentials == 1 && ret >= 0) {
        return ret;
    }
    else if (numPotentials > 1) {
        
        fprintf(stderr, "A relevant candidate must be chosen by lot under Schedule 4, part 4.2, clause 8, section 2 (b): \n");
        potentials[0] = ret;
        for (i = 0 ; i < numPotentials ; i++) {
            fprintf(stderr, "%d: candidate %d\n", i, potentials[i]);            
        }
        
        return potentials[getIntegerFromUser(0, numPotentials-1)];
    }
    
    /* An error has occured somewhere*/
    fprintf(stderr, "An error occured when picking a candidate to exclude!\n");
    return -1;
}

/******************************************************************************/
/*  See schedule 4, Part 4.2, Clause 8, (1,2) */
int pickExcludedCandToDist(electionDef *myE) {
    int i;        
    
    /* There should only be one 'justexcluded' candidate at any one time, so this loop finds them */
    for (i = 0; i < myE->totalCandidateCount ; i++) {
        
        if (myE->candidates[i].status == justexcluded) {
            
            return i;
            
        }
    }
    
    /*Something wrong */
    return -1;
}


/******************************************************************************/
/*  See schedule 4, Part 4.2, Clause 7, (2,3) */
int pickElectedCandToDist(electionDef *myE) {
    int i;     
    int ret = -1;
    int potentials[MAXCAND];
    int numPotentials = 1;
    int upToCount;
        
    // printf("Pick an elected candidate\n");
    
    for (i = 0; i < myE->totalCandidateCount ; i++) {

        if (myE->candidates[i].status == justelected) {
            
            if (ret == -1) {
                ret = i;                
            } else if (myE->candidates[i].statusObtainedAtCount < myE->candidates[ret].statusObtainedAtCount) {
                /* If quota obtained at earliest count  s (3)(a)*/                
                ret = i;
            } else if (myE->candidates[i].statusObtainedAtCount == myE->candidates[ret].statusObtainedAtCount) {
                /*Same count number to obtain quota: (3)(b)*/
                
                if (myE->candidates[i].surplus > myE->candidates[ret].surplus) {
                    /*if surplus is more (3)(b)*/
                    ret = i;
                }
                else if (myE->candidates[i].surplus == myE->candidates[ret].surplus) {
                    /* Same surplus, check total votes back through counts: (3)(c) */

                    upToCount = myE->countNum;
                    
                    /* Do countback here  (3)(c)(i)*/
                    while (voteCountUpTo(&myE->candidates[i], upToCount) == 
                           voteCountUpTo(&myE->candidates[ret], upToCount) && upToCount >= 0) {
                        
                        upToCount --;
                    }
                    
                    if (voteCountUpTo(&myE->candidates[i], upToCount) == 
                        voteCountUpTo(&myE->candidates[ret], upToCount)) {
                        /* No difference in votes when counting back, so add this new voter as a potential
                        for the commissioner to choose by lot (3)(c)(ii) */
                    
                        potentials[numPotentials] = i;
                        numPotentials++;
                    } else if (voteCountUpTo(&myE->candidates[i], upToCount) > 
                               voteCountUpTo(&myE->candidates[ret], upToCount)) {
                        /* this candidate had more at the first unequal count - so they are now the one to be elected */
                        ret = i;
                    }
                }
                
            }

        }
    }        
     
    if (numPotentials == 1 && ret >= 0) {
        return ret;
    }
    else if (numPotentials > 1) {
     
        fprintf(stderr, "A relevant candidate must be chosen by lot under Schedule 4, part 4.2, clause 7, section 3 (c)(ii): \n");
        potentials[0] = ret;
        for (i = 0 ; i < numPotentials ; i++) {
            fprintf(stderr, "%d: candidate %d\n", i, potentials[i]);            
        }
        
        return potentials[getIntegerFromUser(0, numPotentials-1)];
    }
        
    /* An error has occured */
    printf("An error occured when selecting the next candidate to distribute (elected candidates)\n");
    return -1;
}

/******************************************************************************/
/*  See Schedule 4, Pt 4.2, clause 4 (1-2) */
int ceaseScrutiny(electionDef *myE) {
    int i;        
    
    if (num_remaining_cand(myE) <=
        num_remaining_seats(myE) &&
        isSomeoneJustElected(myE) == 0 &&
        isSomeoneJustExcluded(myE) == 0) {
                  
        /* sort candidates by pref count */
        sortByPrefs(myE, 1);
        
        for(i = 0; i < myE->totalCandidateCount ; i++) {
            if (myE->candidates[myE->orderedCandidates[i].index].status == limbo) {
                electCandidate(myE,
                               myE->orderedCandidates[i].index);
            }
        }

        return 1;
        
    }

    return 0;
    
}

/******************************************************************************/
int isSomeoneJustElected ( electionDef *myE) {
    int i;
    int count = 0;

    for (i = 0 ; i < myE->totalCandidateCount ; i++) {
        if (myE->candidates[i].status == justelected) {
            count++;
        }
    }

    return count;
}

/******************************************************************************/
int isSomeoneJustExcluded ( electionDef *myE) {
    int i;
    int count = 0;

    for (i = 0 ; i < myE->totalCandidateCount ; i++) {
        if (myE->candidates[i].status == justexcluded) {
            count++;
        }
    }

    return count;
}

/******************************************************************************/
int num_remaining_cand ( electionDef *myE) {
    int i;
    int count = 0;
        
    for (i = 0 ; i < myE->totalCandidateCount ; i++) {
        if (myE->candidates[i].status == limbo) {
            count++;
        }
    }

    return count;
    
}

/******************************************************************************/
int num_remaining_seats ( electionDef *myE) {

    return (myE->vacancyCount - myE->nextElectedCandidate);

}


/******************************************************************************/
int electCandidate( electionDef *myE, int thisWinner) {
   // printf("Elect %d (%s) as %d\n", thisWinner, myE->candidates[thisWinner].name, myE->nextElectedCandidate +1);
    
    myE->electedCandidates[myE->nextElectedCandidate] = thisWinner;
    myE->nextElectedCandidate ++;
    myE->candidates[thisWinner].status = justelected;
    myE->candidates[thisWinner].statusObtainedAtCount = myE->countNum;
    myE->candidates[thisWinner].surplus = voteCount(&(myE->candidates[thisWinner])) - myE->quota;
    
    if (myE->nextElectedCandidate >= myE->vacancyCount) {       
        return 1;
    }
    return 0;
}


/******************************************************************************/
void distribute_elect ( electionDef *myE, int thisWinner, int surplus) {
    int i, count, j;
    int distCount[MAXCAND+1];
    tValDef rate;
    weightedVoteDef *candVotes;
    int prevRound;
    
    count = 0;
    for (i = 0 ; i < MAXCAND+1 ; i++) {
        distCount[i] = 0;
    }

    /* This is a bit messy - we do 3 passes through the ballots.  
        First we find the most recent count when the candidate recieved ballots,
        then we count the ballots to be distributed (so we can calc. the rate),
        then we do another pass to actually distribute the votes.    
    */
    
    /* First pass - Find the most recent 'count' that gave us some ballots */
    candVotes = myE->candidates[thisWinner].voteCount;
    prevRound = myE->countNum - 1;
    
    while (candVotes != EMPTY_LIST ) {
        
        if (candVotes->countNum == prevRound)  {
            
            if (candVotes->numVotes == 0) {
                prevRound--;
                candVotes = myE->candidates[thisWinner].voteCount;
                continue;
            }
            else {
                count = candVotes->numVotes;
                break;
            }
            
        }
        
        candVotes = candVotes->next;
        
        if (candVotes == EMPTY_LIST && prevRound > 1) {
            prevRound--;
            candVotes = myE->candidates[thisWinner].voteCount;
        }
        
    }
    
    if (count == 0) {
        /* problem - absolutely no votes what-so-ever.  How did they get elected then? */
        return;
    }

    
    /*  Second pass - count the ballots to be excluded */
    count = 0;
    
    candVotes = myE->candidates[thisWinner].voteCount;
    
    while (candVotes != EMPTY_LIST) {
        
        int candNum;
        ballotList testBallot;
        ballotList tmpBallot;
        
        if (candVotes->countNum == prevRound)  {
        
            if (candVotes->numVotes == 0) {
                prevRound--;
                candVotes = myE->candidates[thisWinner].voteCount;
                continue;
            }
            
            testBallot = candVotes->votelist;
        
            while (testBallot != EMPTY_LIST) {
                
                candNum = candParty2Num(myE,
                                        testBallot->orderedVotes[2]);
                
                j = 3;
                while (candNum != -1 && myE->candidates[candNum].status != limbo) {
                    candNum = candParty2Num(myE,
                                            testBallot->orderedVotes[j]);
                    j++;
                }
                
                tmpBallot = testBallot;
                testBallot = testBallot->next;
                
                if (candNum != -1) {
                    count++;
                }
                else {
                    distCount[MAXCAND]++;                
                    count++;
                }
            }
            break;
        }
        candVotes = candVotes->next;
        
    }
    
    
    /*  Calculate value rate */
    
  //  printf("Distribute %d votes:\n", count);
    
    rate.numerator = (surplus);
    rate.denominator = (count - distCount[MAXCAND]);
        
    /* But if the rate is more than the rate of the ballots we have been given, use that instead */
    /* Pt 4.1, clause 1, s 2 */
    if (tVal_gt(rate,candVotes->tVal)) {
        rate.numerator = candVotes->tVal.numerator;
        rate.denominator = candVotes->tVal.denominator;
    }
    
  //  printf("Rate %f  (%d / %d)\n", rate, surplus, count - distCount[MAXCAND]);
    
    /* Third pass - count the votes for each remaining candidate */
    
    candVotes = myE->candidates[thisWinner].voteCount;
    count = 0;
     
    while (candVotes != EMPTY_LIST) {
        
        int candNum;
        ballotList testBallot;
        ballotList tmpBallot;

        if (candVotes->countNum == prevRound)  {

            testBallot = candVotes->votelist;
            
            while (testBallot != EMPTY_LIST) {
            
                candNum = candParty2Num(myE,
                                        testBallot->orderedVotes[2]);
                
                j = 3;
                while (candNum != -1 && myE->candidates[candNum].status != limbo) {
                    candNum = candParty2Num(myE,
                                            testBallot->orderedVotes[j]);
                    j++;
                }
                
                tmpBallot = testBallot;
                testBallot = testBallot->next;
                
                if (candNum != -1) {
                    
                    tmpBallot->next = NULL;
                    
                    add_vote(myE, candNum, 1, rate, tmpBallot, myE->countNum);
                    
                    // printf("Give 1 vote to %d \n", candNum);
                    distCount[candNum]++;
                    
                    count++;
                } else {
                    /* ballot is exhausted */
                    tmpBallot->next = myE->exhaustedBallots;
                    myE->exhaustedBallots = tmpBallot;
                    
                    count++;                
                }
            }
            
            candVotes->votelist = NULL;            
            break;
        }
        
        candVotes = candVotes->next;
        
    }
    
    printRound(myE, myE->countNum, distCount, distCount[MAXCAND], rate);

    /*
    for (i = 0 ; i < myE->totalCandidateCount ; i++) {

        printf("Transfer %d * %f = %d to Candidate %d\n", distCount[i],
               rate, (int)(distCount[i] * rate), i);

    }
    */

}

/******************************************************************************/
int distribute_exclude ( electionDef *myE, int thisWinner) {
    int i, count, j;
    int distCount[MAXCAND+1];
  //float rate;
    weightedVoteDef *candVotes;
    tValDef rate;
    
    rate.numerator = -1;
    rate.denominator = 1;
    
    count = 0;

    for (i = 0 ; i < MAXCAND+1; i++) {
        distCount[i] = 0;   
    }
    

   // printf("Exluding %d\n", thisWinner);
    
    /* Find highest group transfer value */
    candVotes = myE->candidates[thisWinner].voteCount;    
    while (candVotes != EMPTY_LIST) {
        if (tVal_gt(candVotes->tVal,rate)) {
            rate.numerator = candVotes->tVal.numerator;
            rate.denominator = candVotes->tVal.denominator;
        }
        candVotes = candVotes->next;
    }
    
    /* If didn't find any new groups of ballots, then we are done with this exclusion */
    if (tVal_float(rate) == -1) {
        return 0;
    }
        
    candVotes = myE->candidates[thisWinner].voteCount;
    while (candVotes != EMPTY_LIST) {
    
        int candNum;
        ballotList testBallot;
        ballotList tmpBallot;

        if (tVal_float(candVotes->tVal) == -1) {
            //          printf("Skip bc -1\n");
            candVotes = candVotes->next;
            continue;
        }
        
        if (! tVal_e(rate,candVotes->tVal)) {
            //        printf("Skip bc not thisRate\n");
            candVotes = candVotes->next;
            continue;
        }
        
        testBallot = candVotes->votelist;

        while (testBallot != EMPTY_LIST) {

            candNum = candParty2Num(myE,
                                    testBallot->orderedVotes[2]);

            j = 3;
            while (candNum != -1 && myE->candidates[candNum].status != limbo) {
                    
                     candNum = candParty2Num(myE,
                                            testBallot->orderedVotes[j]);
                    j++;
            }

            tmpBallot = testBallot;
            testBallot = testBallot->next;
                                
            if (candNum != -1) {

                    tmpBallot->next = NULL;
                    
                    add_vote(myE, candNum, 1, candVotes->tVal, tmpBallot, myE->countNum);
                   // printf("Give 1 vote to %d \n", candNum);                    
                    distCount[candNum]++;
                    
                    count++;
            } else {
                    
                distCount[MAXCAND]++;
                //Exhaust the vote 
                tmpBallot->next = myE->exhaustedBallots;
                myE->exhaustedBallots = tmpBallot;
                
                count++;                
            }
        }
         
        candVotes->votelist = NULL;
        
        candVotes->tVal.numerator = -2;
        candVotes->tVal.denominator = 1;
        candVotes = candVotes->next;
    }
    
    printRound(myE, myE->countNum, distCount, distCount[MAXCAND], rate);
         
    rate.numerator = -1;
    rate.denominator = 1;
    
    /* Check if any more groups of ballots exist for this candidate that need
        to be distributed */    
    candVotes = myE->candidates[thisWinner].voteCount;    
    while (candVotes != EMPTY_LIST) {
        if (tVal_gt(candVotes->tVal,rate)) {
            rate.numerator = candVotes->tVal.numerator;
            rate.denominator = candVotes->tVal.denominator;
        }
        candVotes = candVotes->next;
    }
    
    /* If didn't find any new groups of ballots, then we are done with this exclusion */
    if (tVal_float(rate) == -1) {
        return 0;
    }
    
    /* Otherwise, we have more ballots to count, so don't finish this exclusion */
    return 1;
    
}

/******************************************************************************/
/* Given a voteDef with party code and cand code within the party, return the unique candidate index number */
int candParty2Num( electionDef *myE, voteDef name) {
    if (name.pcode >= 0 && name.ccode >= 0) {
        return myE->parties[name.pcode].candidates[name.ccode];
    } else {
        return -1;
    }
}


/************************************************************
  Utility functions...
*************************************************************/
void sortByPrefs(electionDef *myE, int prefNum) {
#pragma unused(prefNum)
    // reset ordered array.
    memcpy (myE->orderedCandidates, myE->candidates, myE->totalCandidateCount * sizeof(candidateDef));
    qsort (myE->orderedCandidates, myE->totalCandidateCount, sizeof(candidateDef), &sortPrefs);
}

int sortPrefs (const void *in1, const void *in2) {
    candidateDef *c1 = (candidateDef *) in1;
    candidateDef *c2 = (candidateDef *) in2;
    
    if (voteCount(c1) > voteCount(c2)) {
        return -1;
    }
    else if (voteCount(c1) < voteCount(c2)) {
        return 1;
    }
    return 0;
}


void freeElection(electionDef *myE) {
    int i;
    ballotDef *tempBallot, *nextB;
    weightedVoteDef *vc, *next;
    
    for (i = 0; i < myE->totalCandidateCount; i++) {
        vc = myE->candidates[i].voteCount;

      //  printf("Free candidate %d\n", i);
        while (vc != NULL) {
            tempBallot = vc->votelist;
            while (tempBallot != NULL) {
                nextB = tempBallot->next;
                free(tempBallot);
                tempBallot = nextB;
            }
            next = vc->next;
            free(vc);
            vc = next;            
        }
    }
    
  //  printf("Free exhausted ballots\n");
    tempBallot = myE->exhaustedBallots;
    
    while (tempBallot != NULL) {
        nextB = tempBallot->next;
        free(tempBallot);
        tempBallot = nextB;
    }
        
    for (i = 0 ; i < myE->totalPartyCount ; i++) {
        free(myE->parties[i].candidates);
    }
    
  //  printf("Free other stuff\n");
    free(myE->parties);
    free(myE->candidates);
    free(myE->orderedCandidates);
    free(myE->alphaCandidates);
    free(myE->electorates);
    free(myE->electedCandidates);
    free(myE->ballotfile);
    free(myE->candfile);
    free(myE->partyfile);
    
    myE->parties = NULL;
    myE->candidates = NULL;
    myE->orderedCandidates = NULL;
    myE->alphaCandidates = NULL;
    myE->electorates = NULL;
    myE->electedCandidates  = NULL;
    myE->ballotfile = NULL;
    myE->candfile = NULL;
    myE->partyfile = NULL;
}

